/* expand.cc: general expansion.  Most of this used to reside in
   expand.c.  However, I replaced the GPL'd code with C++ code.
   ccr@mupad.de, 25/9/1998

Copyright (C) 1993, 94, 95, 96, 97 Karl Berry & O. Weber.

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <math.h>
extern "C" {
#include <kpathsea/config.h>

#include <kpathsea/c-pathch.h>
#include <kpathsea/expand.h>
#include <kpathsea/pathsearch.h>
#include <kpathsea/tilde.h>
#include <kpathsea/variable.h>
#include <kpathsea/concatn.h>
#include <kpathsea/absolute.h>
	   }

#include <list>

/* gcc 2.7.* can't handle namespaces. VC++ wants them 
 * and doesn't like using declarations. 
 * 'using namespace std' can't work, since there's std::string,
 * which collides with kpathsea's typedef of string to mean char*. */
#if defined(WIN32) || (defined(__GNUC__) && __GNUC__ > 2)
#define list std::list
#define copy std::copy
#define back_inserter std::back_inserter
#define swap std::swap
#endif

/* Do variable expansion first so ~${USER} works.  (Besides, it's what the
   shells do.)  */

extern "C" string
kpse_expand P1C(const_string, s)
{
  string var_expansion = kpse_var_expand (s);
  string tilde_expansion = kpse_tilde_expand (var_expansion);
  
  /* `kpse_var_expand' always gives us new memory; `kpse_tilde_expand'
     doesn't, necessarily.  So be careful that we don't free what we are
     about to return.  */
  if (tilde_expansion != var_expansion)
    free (var_expansion);
  
  return tilde_expansion;
}


/* Forward declarations of functions as defined below */
static char **brace_expand P1H(const_string);
static void free_array P1H(char **);

/* If $KPSE_DOT is defined in the environment, prepend it to any relative
   path components. */

static string
kpse_expand_kpse_dot P1C(string, path)
{
  string ret, elt;
  string kpse_dot = getenv("KPSE_DOT");
#ifdef MSDOS
  boolean malloced_kpse_dot = false;
#endif
  
  if (kpse_dot == NULL)
    return path;
  ret = (string)kp_xmalloc(1);
  *ret = 0;

#ifdef MSDOS
  /* Some setups of ported Bash force $KPSE_DOT to have the //d/foo/bar
     form (when `pwd' is used), which is not understood by libc and the OS.
     Convert them back to the usual d:/foo/bar form.  */
  if (kpse_dot[0] == '/' && kpse_dot[1] == '/'
      && kpse_dot[2] >= 'A' && kpse_dot[2] <= 'z' && kpse_dot[3] == '/') {
    kpse_dot++;
    kpse_dot = xstrdup (kpse_dot);
    kpse_dot[0] = kpse_dot[1];  /* drive letter */
    kpse_dot[1] = ':';
    malloced_kpse_dot = true;
  }
#endif

  for (elt = kpse_path_element (path); elt; elt = kpse_path_element (NULL)) {
    string save_ret = ret;
    /* We assume that the !! magic is only used on absolute components.
       Single "." get special treatment, as does "./" or its equivalent. */
    if (kpse_absolute_p (elt, false) || (elt[0] == '!' && elt[1] == '!')) {
      ret = concat3(ret, elt, ENV_SEP_STRING);
    } else if (elt[0] == '.' && elt[1] == 0) {
      ret = concat3 (ret, kpse_dot, ENV_SEP_STRING);
#ifndef VMS
    } else if (elt[0] == '.' && IS_DIR_SEP(elt[1])) {
      ret = concatn (ret, kpse_dot, elt + 1, ENV_SEP_STRING, NULL);
    } else {
      ret = concatn (ret, kpse_dot, DIR_SEP_STRING, elt, ENV_SEP_STRING, NULL);
#endif
    }
    free (save_ret);
  }

#ifdef MSDOS
  if (malloced_kpse_dot) free (kpse_dot);
#endif

  ret[strlen (ret) - 1] = 0;
  return ret;
}

/* Do brace expansion on ELT; then do variable and ~ expansion on each
   element of the result; then do brace expansion again, in case a
   variable definition contained braces (e.g., $TEXMF).  Return a
   string comprising all of the results separated by ENV_SEP_STRING.  */

static string
kpse_brace_expand_element P1C(const_string, elt)
{
  unsigned i;
  string *expansions = brace_expand (elt);
  string ret = (string)kp_xmalloc (1);
  *ret = 0;

  for (i = 0; expansions[i]; i++) {
    /* Do $ and ~ expansion on each element.  */
    string x = kpse_expand (expansions[i]);
    string save_ret = ret;
    if (!STREQ (x, expansions[i])) {
      /* If we did any expansions, do brace expansion again.  Since
         recursive variable definitions are not allowed, this recursion
         must terminate.  (In practice, it's unlikely there will ever be
         more than one level of recursion.)  */
      string save_x = x;
      x = kpse_brace_expand_element (x);
      free (save_x);
    }
    ret = concat3 (ret, x, ENV_SEP_STRING);
    free (save_ret);
    free (x);
  }

  free_array (expansions);
  ret[strlen (ret) - 1] = 0; /* waste the trailing null */
  return ret;
}

/* Be careful to not waste all the memory we allocate for each element.  */

extern "C" string
kpse_brace_expand P1C(const_string, path)
{
  string kpse_dot_expansion;
  string elt;
  unsigned len;
  /* Must do variable expansion first because if we have
       foo = .:~
       TEXINPUTS = $foo
     we want to end up with TEXINPUTS = .:/home/karl.
     Since kpse_path_element is not reentrant, we must get all
     the path elements before we start the loop.  */
  string xpath = kpse_var_expand (path);
  string ret = (string)kp_xmalloc (1);
  *ret = 0;

  for (elt = kpse_path_element (xpath); elt; elt = kpse_path_element (NULL)) {
    string save_ret = ret;
    /* Do brace expansion first, so tilde expansion happens in {~ka,~kb}.  */
    string expansion = kpse_brace_expand_element (elt);
    ret = concat3 (ret, expansion, ENV_SEP_STRING);
    free (expansion);
    free (save_ret);
  }

  /* Waste the last byte by overwriting the trailing env_sep with a null.  */
  len = strlen (ret);
  if (len != 0)
    ret[len - 1] = 0;
  free (xpath);

  kpse_dot_expansion = kpse_expand_kpse_dot (ret);
  if (kpse_dot_expansion != ret)
    free (ret);

  return kpse_dot_expansion;
}

/* Expand all special constructs in a path, and include only the actually
   existing directories in the result. */
extern "C" string
kpse_path_expand P1C(const_string, path)
{
  string ret;
  string xpath;
  string elt;
  unsigned len;

  /* Initialise ret to the empty string. */
  ret = (string)kp_xmalloc (1);
  *ret = 0;
  len = 0;
  
  /* Expand variables and braces first.  */
  xpath = kpse_brace_expand (path);

  /* Now expand each of the path elements, printing the results */
  for (elt = kpse_path_element (xpath); elt; elt = kpse_path_element (NULL)) {
    str_llist_type *dirs;

    /* Skip and ignore magic leading chars.  */
    if (*elt == '!' && *(elt + 1) == '!')
      elt += 2;

    /* Do not touch the device if present */
    if (NAME_BEGINS_WITH_DEVICE (elt)) {
      while (IS_DIR_SEP (*(elt + 2)) && IS_DIR_SEP (*(elt + 3))) {
        *(elt + 2) = *(elt + 1);
        *(elt + 1) = *elt;
        elt++;
      }
    } else {
      /* We never want to search the whole disk.  */
      while (IS_DIR_SEP (*elt) && IS_DIR_SEP (*(elt + 1)))
        elt++;
    }

    /* Search the disk for all dirs in the component specified.
       Be faster to check the database, but this is more reliable.  */
    dirs = kpse_element_dirs (elt); 
    if (dirs && *dirs) {
      str_llist_elt_type *dir;

      for (dir = *dirs; dir; dir = STR_LLIST_NEXT (*dir)) {
        string thedir = STR_LLIST (*dir);
        unsigned dirlen = strlen (thedir);
        string save_ret = ret;
        /* Retain trailing slash if that's the root directory.  */
        if (dirlen == 1 || (dirlen == 3 && NAME_BEGINS_WITH_DEVICE (thedir)
                            && IS_DIR_SEP (thedir[2]))) {
          ret = concat3 (ret, thedir, ENV_SEP_STRING);
          len += dirlen + 1;
          ret[len - 1] = ENV_SEP;
        } else {
          ret = concat (ret, thedir);
          len += dirlen;
          ret [len - 1] = ENV_SEP;
        }
        free (save_ret);
      }
    }
  }
  /* Get rid of trailing ':', if any. */
  if (len != 0)
    ret[len - 1] = 0;
  return ret;
}

/* Here, the Bash-Code was to be found. */

/* The replacement code needs the following functions:

static char **brace_expand P1H(const_string);
static void free_array P1H(char **);

*/

/* Starting at a '{', find the matching '}' */
static const char *find_matching_brace(const_string fromhere) {
  int level=1;
  const char *s=fromhere;
  s++;
  while(*s && level > 0) {
    switch (*s) {
    case '{':
      ++level; 
      break;
    case '}':
      --level;
      break;
    case '\\':
      s++;
      break;
    default:
      break;
    }
    if(*s && level>0) /* might have been increased by a \\ */
      s++;
  }
  if(level>0)
    return NULL;
  else
    return s;
}

/* count the non-escaped occurences of a character x in a substring,
   not counting those inside nested braces. */
int countn(const_string str, int len, const char x) {
  int n=0,level=0,i;
  for(i=0; i<len && str[i]; i++) {
    if(str[i]=='{')
      ++level;
    if(str[i]=='}')
      --level;
    if(level==0 && x == str[i])
      n++;
    if(str[i] == '\\')
      i++;
  }
  return n;
}

/* find the first non-escaped occurence of char x in substring,
   skipping nested braces. */
static const char *nonesc_index(const_string str, int maxlen, const char x) {
  int i, level=0;
  for(i=0; i<maxlen && str[i]; i++) {
    if(level == 0 && x == str[i])
      return(str+i);
    if(str[i]=='{')
      ++level;
    if(str[i]=='}')
      --level;
    if('\\' == str[i])
      i++;
  }
  return NULL;
}

char *strndup(const_string input, unsigned int len) {
  char *ans = (char *)malloc((len+1)*sizeof(char));
  strncpy(ans, input, len);
  ans[len]='\0';
  return(ans);
}

/* Split the input string at the given character, leaving nested braces
   alone. */
static list<char *> splitstring(const_string input, 
				const int maxlen, const char split) {
  list<char *> ans;
  const char *s, *t;
  int len=maxlen;

  s=input;
  t=nonesc_index(s,len,split);
  while(t && *t && *s) {
    ans.push_back(strndup(s,t-s));
    len -= (t-s+1);
    s = t+1;
    t = nonesc_index(s,len,split);
  }
  if(s && *s && len>0) {
    ans.push_back(strndup(s,len));
  }
  return ans;
}

/* Expand alternatives in curly braces. */
static list<char *> _brace_expand(const_string input) {
  list<char *> ans;
  const char *brace, *braceend, *prebrace, *postbrace;

  if((brace=nonesc_index(input,strlen(input),'{')) == NULL) {
    ans.push_back(strdup(input));
  } else {
    prebrace=strndup(input, brace-input);
    braceend=find_matching_brace(brace);
    if(NULL==braceend) {
		free((void*)prebrace);
		/* xxx error handling xxx */
		return ans;
    }
    /* braceend points to a closing brace, so braceend+1 is safe. */
    postbrace=strdup(braceend+1); 
    brace++;

    list<char *> parts=splitstring(brace, braceend-brace,
				   BRACE_ARG_SEP);
    if(parts.size()==1) {
      free(*(parts.begin()));
//      parts.clear();
// list<...>::clear() is missing on the old local.
      parts.erase(parts.begin());
      parts=splitstring(brace, braceend-brace, ENV_SEP);
    }
    list<char *>::iterator curstring;
    /* Run each substring through the expansion again. */
    for(curstring = parts.begin(); curstring != parts.end(); ++curstring) {
      list<char *> tmp = _brace_expand(*curstring);
	  copy(tmp.begin(), tmp.end(), back_inserter(ans));
      free(*curstring);
    }
    parts.erase(parts.begin(),parts.end());

    /* Now, combine those strings with the stuff in front of the
       braces and behind the braces. */
    for(curstring = ans.begin(); curstring != ans.end();
	++curstring) {
      char *tmp = (char *) kp_xmalloc(strlen(prebrace)+strlen(*curstring)
				   +strlen(postbrace)+1);
      strcpy(tmp,prebrace);
      strcat(tmp,*curstring);
      strcat(tmp,postbrace);
      free(*curstring);
      *curstring=tmp;
    }

	free((void*)prebrace);
	free((void*)postbrace);

    /* Finally, yet another recursive call. */
    swap(parts,ans);

    for(curstring = parts.begin(); curstring != parts.end(); ++curstring) {
      list<char *> tmp = _brace_expand(*curstring);
	  copy(tmp.begin(), tmp.end(), back_inserter(ans));
      free(*curstring);
    }
    parts.erase(parts.begin(),parts.end());
  }

  return(ans);
}

static char **brace_expand (const_string input) {
  list<char *> parts = _brace_expand(input);
  list<char *>::iterator pos;
  char **ans = (char**) kp_xmalloc((parts.size()+1)*sizeof(char*));
  int i;

  for(i=0, pos=parts.begin(); 
      pos != parts.end(); ++pos) {
    ans[i++] = *pos;
  }
  ans[parts.size()]=NULL;
  return(ans);
}


/* Free a NULL-terminated array of char* */
static void free_array (char ** array) {
  int i;
  
  if(NULL==array)
    return;

  for(i=0; NULL!=array[i]; i++)
    free(array[i]);

  free(array);
}
